/*
 *
 *
 */


#include <errno.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/time.h>

#include <security/pam_modules.h>
#include <security/pam_modutil.h>
#include <security/pam_ext.h>
#include <security/_pam_macros.h>
#include <curl/curl.h>

#include "cJSON.h"
#include "base64.h"


#define STRLEN 128
#define STRNUM	1024
#define CONFDIR	"/etc/yyssh"
#define USERCONFILE "pam_yyssh.conf"
#define USERDATA "pam_record.dat"
#define CONFILE	CONFDIR"/"USERCONFILE
#define MODE_EVERY_TIME	1	/* Authciaction every times */
#define MODE_SMOOTH		2	/* Will be triggerd according to policy */
#define MODE_DISABLE	3	/* Disable dynamic code verify. */
#define MODE_DEFAULT	MODE_EVERY_TIME

#define POLICY_CLIENT	0x0010
#define POLICY_SRCIP	0x0020
#define	POLICY_DAWN		0x0040
#define	POLICY_SET (POLICY_CLIENT|POLICY_SRCIP|POLICY_DAWN)

#define PROMPT_STRING "[YY安全中心(YY-Security-Center)] verification code: "

/*Define json key name */
#define JSON_RHOST	"rhost"
#define JSON_CLIENT	"client"
#define JOSN_DAWN	"dawn"
#define	JSON_TIMESTAMP	"timestamp"
#define JSON_RECORD	"record"
#define JSON_AUTH_TIME	"auth_time"

#define check_substring_avaiable(ptr) ({    \
		if(ptr == NULL) \
			break; \
		if(ptr[0] == '#' || ptr[0] == '\0') \
			continue;\
		})

#define MIN(a, b)	(((a) < (b))? (a): (b))

static struct yyssh_s {
	pam_handle_t *pamh;
	int mode;
	bool debug;
	int smooth_policy;
	int policy_expire;
	int dawn_time_range_min;
	int dawn_time_range_max;
	int dawn_time_expire;

	bool alias_enable;
	char *username_alias;
	
	int appid;
	char *appkey;
	char *user;
	char *sftoken;
	char *request_url;

	bool proxy;
	char *proxy_user;
	char *proxy_passwd;
	char *proxy_url;
	char *help_url;

	long int updatetimestamp;
	cJSON *ucj;
	struct timeval tv;
	struct tm tm;
	bool update_record;

	char *remote_version;
	char *rhost;
	cJSON *cj;
	char *res;
} yyssh_t;
static struct yyssh_s yyssh;

static int mode_vaild(int mode)
{
	return (mode != MODE_EVERY_TIME && mode != MODE_SMOOTH && mode != MODE_DISABLE)? -1: 1;
}

char *get_string(char *string)
{
	char *str, *token, *saveptr;

	for(str = string; ; str = NULL) {

		token = strtok_r(str, "\t ", &saveptr);
		if(token == NULL || token[0] == '#')
			return NULL;

		return token;
	}

	return NULL;
}

long int get_integer(char *string)
{
	char *str, *token, *saveptr;

	for(str = string; ; str = NULL) {

		token = strtok_r(str, "\t ", &saveptr);
		if(token == NULL || token[0] == '#')
			return -1;

		return atol(token);
	}

	return 0;
}

static int load_global_configure(pam_handle_t *pamh)
{
	int fd, n;
	struct stat st;
	time_t timep;
	char *T;
	char *confbuff, *str, *token, *saveptr1, *substr, *saveptr2, *subtoken;
	char *saveptr3, *substr1, *subtoken1;

	if(access(CONFILE, O_RDONLY) < 0) {

	  	pam_syslog(yyssh.pamh, LOG_DEBUG, "access file %s fail.", CONFILE);
		return PAM_SYSTEM_ERR;
	}

	memset(&st, 0, sizeof(struct stat));
	if(stat(CONFILE, &st) != 0)
		return PAM_SYSTEM_ERR;

	fd = open(CONFILE, O_RDONLY);
	if(fd < 0)
		return PAM_SYSTEM_ERR;

	confbuff = calloc(1, st.st_size+1);
	if(confbuff == NULL) {
		pam_syslog(yyssh.pamh, LOG_DEBUG, 
				"alloc confbuff fail, size:%d bytes.", (int)st.st_size);
		return PAM_BUF_ERR;
	}

	n = read(fd, confbuff, st.st_size);
	if(n < st.st_size) {
		pam_syslog(yyssh.pamh, LOG_DEBUG, 
				"read configure size:%d less than file size:%d", n, (int)st.st_size);
		return PAM_SYSTEM_ERR;
	}
	
	for(str = confbuff;;str = NULL) { 
		token = strtok_r(str, "\n", &saveptr1);
		check_substring_avaiable(token);

		for(substr = token;;substr = NULL) {
			subtoken = strtok_r(substr, "\t ", &saveptr2);
			if(subtoken == NULL)
				break;


			if((strcmp("mode", subtoken)) == 0) {
				int mm = (int)get_integer(saveptr2);
				if(mode_vaild(mm) > 0) {
					yyssh.mode = mm;
				} else {
					pam_syslog(yyssh.pamh, LOG_DEBUG, 
							"read configure mode:%d not vaild, and set %d by default.", 
							mm, MODE_DEFAULT);
					yyssh.mode = MODE_DEFAULT;
				}

			}else if(strcmp("debug", subtoken) == 0) {
				T = get_string(saveptr2);
				yyssh.debug = (strcmp(T, "true") == 0)? true: false;

			}else if(strcmp("appid", subtoken) == 0) {
				yyssh.appid = (int)get_integer(saveptr2);
				
			}else if(strcmp("appkey", subtoken) == 0) {
				yyssh.appkey = get_string(saveptr2);

			}else if(strcmp("request_url", subtoken) == 0){
				yyssh.request_url = get_string(saveptr2);

			}else if(strcmp("proxy", subtoken) == 0) {
				T = get_string(saveptr2);
				yyssh.proxy = (strcmp(T, "true") == 0)? true: false;
				
			}else if(strcmp("proxy_user", subtoken) == 0) {
			
				yyssh.proxy_user = get_string(saveptr2);
			}else if(strcmp("proxy_passwd", subtoken) == 0) {
				yyssh.proxy_passwd = get_string(saveptr2);
			
			}else if(strcmp("proxy_url", subtoken) == 0) {
				yyssh.proxy_url = get_string(saveptr2);
			
			}else if(strcmp("help_url", subtoken) == 0) {
				yyssh.help_url = get_string(saveptr2);		

			}else if(strcmp("policy_expire", subtoken) == 0) {
				yyssh.policy_expire = (int)get_integer(saveptr2);

			}else if(strcmp("smooth_policy", subtoken) == 0) {
				yyssh.smooth_policy = 0;
				for(substr1 = saveptr2;;substr1 = NULL){
					subtoken1 = strtok_r(substr1, ", ", &saveptr3);
					if(subtoken1 == NULL)
						break;
					if(strcmp("client", subtoken1) == 0){
						yyssh.smooth_policy |= POLICY_CLIENT;
					}else if(strcmp("srcip", subtoken1) == 0){
						yyssh.smooth_policy |= POLICY_SRCIP;
					}else if(strcmp("dawntime", subtoken1) == 0){
						yyssh.smooth_policy |= POLICY_DAWN;
					}
				}
			}else if(strcmp("dawn_time_expire", subtoken) == 0){
				yyssh.dawn_time_expire = (int)get_integer(saveptr2);

			}else if(strcmp("dawn_time_range", subtoken) == 0){
				for(substr1 = saveptr2;;substr1 = NULL){
					subtoken1 = strtok_r(substr1, "~", &saveptr3);
					if(subtoken1 == NULL)
						break;
					yyssh.dawn_time_range_min = (int)get_integer(subtoken1);
					yyssh.dawn_time_range_max = (int)get_integer(saveptr3);
					break;
				}
			}

			break;
		}
	}

	gettimeofday(&yyssh.tv, NULL);
	time(&timep);
	localtime_r(&timep, &yyssh.tm);
	pam_get_item(pamh, PAM_USER, (void *)&yyssh.user);
	pam_get_item(pamh, PAM_XDISPLAY, (void *)&yyssh.remote_version);
	pam_get_item(pamh, PAM_RHOST, (void *)&yyssh.rhost);

	return PAM_SUCCESS;
}

static int load_user_configure(char *user)
{
	char str[STRLEN];
	DIR *D;
	int r, fd, mm;
	struct stat st;
	char *buff, *T;
	char *confbuff, *str1, *token, *saveptr1, *substr, *saveptr2, *subtoken;
	char *saveptr3, *substr1, *subtoken1;

	memset(str, 0, STRLEN);
	snprintf(str, STRLEN, "%s/%s", CONFDIR, user);

	D = opendir(str);
	if(D == NULL && errno == ENOENT) {
		r = mkdir(str, S_IRWXU);
		if(r < 0)
			return -1;
	} else {
		closedir(D);
	}

	memset(str, 0, STRLEN);
	snprintf(str, STRLEN, "%s/%s/%s", CONFDIR, user, USERDATA);
	fd = open(str, O_RDONLY);
	if(fd >= 0) {

		memset(&st, 0, sizeof(struct stat));
		if(stat(str, &st) != 0)
			goto release;

		buff = calloc(1, st.st_size + 1);
		r = read(fd, buff, st.st_size);
		if(r < st.st_size) {
			pam_syslog(yyssh.pamh, LOG_DEBUG, 
					"read %s configure file %d size less than %d.", 
					str, r, (int)st.st_size);
			goto release;
		}

		yyssh.ucj = cJSON_Parse((const char *)buff);

release:
		free(buff);
		close(fd);
	}


	memset(str, 0, STRLEN);
	snprintf(str, STRLEN, "%s/%s/%s", CONFDIR, user, USERCONFILE);
	fd = open(str, O_RDONLY);
	if(fd > 0) {
		memset(&st, 0, sizeof(struct stat));
		if(stat(str, &st) != 0)
			goto release1;

		buff = calloc(1, st.st_size + 1);
		r = read(fd, buff, st.st_size);
		if(r < st.st_size){
			pam_syslog(yyssh.pamh, LOG_DEBUG, 
					"read %s configure file %d size less than %d.", 
					str, r, (int)st.st_size);
			goto release1;
		}

		for(str1 = buff;;str1 = NULL) { 
			token = strtok_r(str1, "\n", &saveptr1);
			check_substring_avaiable(token);

			for(substr = token;;substr = NULL) {
				subtoken = strtok_r(substr, "\t ", &saveptr2);
				if(subtoken == NULL)
					break;


				if((strcmp("mode", subtoken)) == 0) {
					int mm = (int)get_integer(saveptr2);
					if(mode_vaild(mm) > 0) {
						yyssh.mode = mm;
					} else {
						pam_syslog(yyssh.pamh, LOG_DEBUG, 
								"read configure mode:%d not vaild, and not change.", mm);
					}

				}else if(strcmp("policy_expire", subtoken) == 0) {
					yyssh.policy_expire = (int)get_integer(saveptr2);

				}else if(strcmp("smooth_policy", subtoken) == 0) {
					yyssh.smooth_policy = 0;
					for(substr1 = saveptr2;;substr1 = NULL){
						subtoken1 = strtok_r(substr1, ", ", &saveptr3);
						if(subtoken1 == NULL)
							break;
						if(strcmp("client", subtoken1) == 0){
							yyssh.smooth_policy |= POLICY_CLIENT;
						}else if(strcmp("srcip", subtoken1) == 0){
							yyssh.smooth_policy |= POLICY_SRCIP;
						}else if(strcmp("dawntime", subtoken1) == 0){
							yyssh.smooth_policy |= POLICY_DAWN;
						}
					}
				}else if(strcmp("dawn_time_expire", subtoken) == 0){
					yyssh.dawn_time_expire = (int)get_integer(saveptr2);

				}else if(strcmp("dawn_time_range", subtoken) == 0){
					for(substr1 = saveptr2;;substr1 = NULL){
						subtoken1 = strtok_r(substr1, "~", &saveptr3);
						if(subtoken1 == NULL)
							break;
						yyssh.dawn_time_range_min = (int)get_integer(subtoken1);
						yyssh.dawn_time_range_max = (int)get_integer(saveptr3);
						break;
					}
				}else if(strcmp("alias_enable", subtoken) == 0){
					T = get_string(saveptr2);
					yyssh.alias_enable = (strcmp(T, "true") == 0)? true: false;

				}else if(strcmp("username_alias", subtoken) == 0){
					T = get_string(saveptr2);
					yyssh.username_alias = calloc(1, (strlen(T)+1));
					strncpy(yyssh.username_alias, T, strlen(T));
				}

				break;
			}
		}

release1:
		free(buff);
		close(fd);
	}

	return PAM_SUCCESS;
}

static int save_user_data(pam_handle_t *pamh)
{
	int fd, i, dawn;
	long int timestamp, k;
	FILE *f;
	char str[STRLEN];
	char *rhost, *client;
	char *rebuff;
	cJSON *root,*arrayfmt, *fmt, *oldfmt, *subitem;

	memset(str, 0, STRLEN);
	snprintf(str, STRLEN, "%s/%s/%s", CONFDIR, yyssh.user, USERDATA);
	f = fopen(str, "w+");
	if(f == NULL)
		return PAM_SYSTEM_ERR;
	else
		fd = fileno(f);
	
	root=cJSON_CreateObject();

	if(yyssh.updatetimestamp != 0) {
		cJSON_AddItemToObject(root, JSON_AUTH_TIME, cJSON_CreateNumber(yyssh.updatetimestamp));
	}else{
		k = 0;
		if(yyssh.ucj != NULL)
			k = cJSON_GetObjectItem(yyssh.ucj, JSON_AUTH_TIME)->valuedouble;
		cJSON_AddItemToObject(root, JSON_AUTH_TIME, cJSON_CreateNumber(k));
	}

	cJSON_AddItemToObject(root, JSON_RECORD, arrayfmt=cJSON_CreateArray());

	if(yyssh.ucj == NULL)
		goto skip;

	/* Scan old item from json */
	oldfmt = cJSON_GetObjectItem(yyssh.ucj, JSON_RECORD);

	for(i = 0; oldfmt != NULL && i<cJSON_GetArraySize(oldfmt); i++){
		fmt = NULL;
		subitem=cJSON_GetArrayItem(oldfmt, i);
		if(subitem == NULL)
			continue;

		rhost = cJSON_GetObjectItem(subitem, JSON_RHOST)->valuestring;
		client = cJSON_GetObjectItem(subitem, JSON_CLIENT)->valuestring;
		dawn = cJSON_GetObjectItem(subitem, JOSN_DAWN)->valueint;
		timestamp = cJSON_GetObjectItem(subitem, JSON_TIMESTAMP)->valuedouble;

		if((yyssh.tv.tv_sec > timestamp) && ((yyssh.tv.tv_sec - timestamp) > yyssh.policy_expire))
			continue;

		if(rhost == NULL || timestamp == 0)
			continue;

		if(yyssh.rhost != NULL && (strcmp(yyssh.rhost, rhost) == 0) && yyssh.tm.tm_hour == dawn){
			if(yyssh.remote_version == NULL && client == NULL){
				continue;
			}else if(yyssh.remote_version != NULL && client != NULL){
				if(strcmp(yyssh.remote_version, client) == 0)
					continue;
			}
		
		}

		cJSON_AddItemToArray(arrayfmt, fmt=cJSON_CreateObject());
		cJSON_AddStringToObject(fmt, JSON_RHOST, rhost);
		cJSON_AddStringToObject(fmt, JSON_CLIENT, client);
		cJSON_AddNumberToObject(fmt, JOSN_DAWN, dawn);
		cJSON_AddNumberToObject(fmt, JSON_TIMESTAMP, timestamp);
	}

skip:
	cJSON_AddItemToArray(arrayfmt, fmt=cJSON_CreateObject());
	
	if(yyssh.rhost == NULL)
		cJSON_AddStringToObject(fmt, JSON_RHOST, "");
	else
		cJSON_AddStringToObject(fmt, JSON_RHOST, yyssh.rhost);

	if(yyssh.remote_version == NULL)
		cJSON_AddStringToObject(fmt, JSON_CLIENT, "");
	else
		cJSON_AddStringToObject(fmt, JSON_CLIENT, yyssh.remote_version);

	cJSON_AddNumberToObject(fmt, JOSN_DAWN, yyssh.tm.tm_hour);
	cJSON_AddNumberToObject(fmt, JSON_TIMESTAMP, yyssh.tv.tv_sec);


	rebuff = cJSON_PrintUnformatted(root);
	write(fd, rebuff, strlen(rebuff));

	if(rebuff != NULL)
		free(rebuff);

	cJSON_Delete(root);

	close(fd);

	return PAM_SUCCESS;
}

static size_t write_callback(void *contents, size_t size, size_t nmemb, void *userp)
{
	size_t realsize = size * nmemb;
	void *ptr;

	ptr = calloc(1, (realsize + 1));

	memcpy(ptr, contents, realsize);

	yyssh.cj = cJSON_Parse((const char *)ptr);

	free(ptr);

	return realsize;
}

static int send_verify(pam_handle_t *pamh)
{
	int r = PAM_SYSTEM_ERR;
	CURL *curl;
	long int res;
	cJSON *rescj;
	int rescode = -1;
	char str1[STRNUM];
	char str2[STRNUM];
	char *str3;

	memset(str1, 0, STRNUM);
	memset(str2, 0, STRNUM);
	yyssh.update_record = false;

	curl = curl_easy_init();
	if(!curl) {
		pam_syslog(pamh, LOG_DEBUG, "initial libcurl fail.");
		goto out;
	}

	if(yyssh.help_url == NULL)
		r = pam_prompt(pamh, PAM_PROMPT_ECHO_ON, &yyssh.sftoken, "%s", PROMPT_STRING);
	else
		r = pam_prompt(pamh, PAM_PROMPT_ECHO_ON, &yyssh.sftoken, 
				"For help,%s\n%s", yyssh.help_url, PROMPT_STRING);

	if(r != PAM_SUCCESS) {
		pam_syslog(pamh, LOG_DEBUG, "user:%s input dynamic code fail.", yyssh.user);
		goto out;
	}

	if(yyssh.alias_enable && yyssh.username_alias && (strcmp(yyssh.username_alias, "") != 0)) {
		pam_syslog(pamh, LOG_DEBUG, "user:%s with alias name:%s", yyssh.user, yyssh.username_alias);
		snprintf(str1, STRNUM, "%s?appid=%d&appkey=%s&user=%s&sftoken=%s", 
				yyssh.request_url, yyssh.appid, yyssh.appkey, yyssh.username_alias, yyssh.sftoken);
	
	}else{
		snprintf(str1, STRNUM, "%s?appid=%d&appkey=%s&user=dw_%s&sftoken=%s", 
				yyssh.request_url, yyssh.appid, yyssh.appkey, yyssh.user, yyssh.sftoken);
	}

	if(yyssh.proxy) {
		str3 = str2;
		base64_encode_alloc(str1, strlen(str1), &str3);
		snprintf(str1, STRNUM, "%s?GET_QUEUE=%s", yyssh.proxy_url, str3);
		snprintf(str3, STRNUM, "%s:%s", yyssh.proxy_user, yyssh.proxy_passwd);
		curl_easy_setopt(curl, CURLOPT_USERPWD, str3);
	}


	curl_easy_setopt(curl, CURLOPT_DNS_CACHE_TIMEOUT, 2L);
	curl_easy_setopt(curl, CURLOPT_TIMEOUT, 4L);
	curl_easy_setopt(curl, CURLOPT_URL, str1);
	curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, write_callback);
	curl_easy_perform(curl);
	curl_easy_getinfo (curl, CURLINFO_RESPONSE_CODE, &res);
	curl_easy_cleanup(curl);

	rescj = cJSON_GetObjectItem(yyssh.cj, "rescode");
	if(yyssh.cj != NULL && rescj != NULL) {
		rescode = rescj->valueint;
	}else if(yyssh.cj != NULL){
		memset(str2, 0, STRNUM);
		str3 = cJSON_PrintUnformatted(yyssh.cj);
		snprintf(str2, STRNUM, "%s", str3);
		free(str3);
	}

	if(rescode == 200){
		r = PAM_SUCCESS;
		yyssh.updatetimestamp = yyssh.tv.tv_sec;
		yyssh.update_record = true;
	}else{
		r = PAM_SYSTEM_ERR;
	}

	pam_syslog(pamh, LOG_DEBUG, "Send %s verification code:%s," 
			" response http:%ld API:%d %s", yyssh.user, yyssh.sftoken, 
			res, rescode, str2);
out:
	return r;
}

static int check_smooth_policy(pam_handle_t *pamh)
{
	cJSON *array, *subitem;
	int i, flags, dawn;
	char *rhost, *client;
	long int timestamp;

	flags = -1;

	if(yyssh.ucj == NULL)
		goto skip_check;

	array = cJSON_GetObjectItem(yyssh.ucj, JSON_RECORD);

	timestamp = cJSON_GetObjectItem(yyssh.ucj, JSON_AUTH_TIME)->valuedouble;
	if((yyssh.tv.tv_sec > timestamp) && ((yyssh.tv.tv_sec - timestamp) > yyssh.policy_expire))
		return send_verify(pamh);
	
	if(array != NULL){
		for(i = 0, flags = 0; i < cJSON_GetArraySize(array); i++){
			flags = 0;
			subitem=cJSON_GetArrayItem(array,i);
			rhost = cJSON_GetObjectItem(subitem, JSON_RHOST)->valuestring;
			client = cJSON_GetObjectItem(subitem, JSON_CLIENT)->valuestring;
			dawn = cJSON_GetObjectItem(subitem, JOSN_DAWN)->valueint;
			timestamp = cJSON_GetObjectItem(subitem, JSON_TIMESTAMP)->valuedouble;

			if((yyssh.tv.tv_sec >= timestamp) && ((yyssh.tv.tv_sec-timestamp) > yyssh.policy_expire))
				continue;

			if(yyssh.smooth_policy & POLICY_CLIENT) {
				if(yyssh.remote_version == NULL && client == NULL){
					flags |= POLICY_CLIENT;
				}else if(yyssh.remote_version != NULL && client != NULL){
					if(strncmp(yyssh.remote_version, client,  (size_t)MIN(strlen(yyssh.remote_version), strlen(client))) == 0)
						flags |= POLICY_CLIENT;
				}
			}

			if(yyssh.smooth_policy & POLICY_SRCIP) {
				if(strcmp(yyssh.rhost, rhost) == 0)
					flags |= POLICY_SRCIP;
			}
			
			if(yyssh.smooth_policy & POLICY_DAWN) {
				if((yyssh.dawn_time_range_min <=yyssh.tm.tm_hour) && (yyssh.tm.tm_hour <= yyssh.dawn_time_range_max)){
					if((yyssh.tv.tv_sec >= timestamp) && ((yyssh.tv.tv_sec-timestamp) <= yyssh.dawn_time_expire))
						flags |= POLICY_DAWN;
				}else {
					flags |= POLICY_DAWN;
				}
			}

			if(flags == yyssh.smooth_policy) {

				break;
			}
		}
	
	}

skip_check:

	if(flags == yyssh.smooth_policy) {
		pam_syslog(pamh, LOG_DEBUG, "rhost:%s dawn:%d client:%s", rhost, dawn, client);
		pam_syslog(pamh, LOG_DEBUG, "Authentication strategy found %s record, and skip check.", yyssh.user);
		return PAM_SUCCESS;
	}else {
		return send_verify(pamh);
	}
}

static int sftoken_auth(const char *pam_type, pam_handle_t *pamh, 
	int flags, int argc, const char **argv)
{
	int r = PAM_SYSTEM_ERR;
	int mode = -1;

	memset(&yyssh, 0, sizeof(struct yyssh_s));
	r = load_global_configure(pamh);
	if(r !=  PAM_SUCCESS)
		return r;

	load_user_configure(yyssh.user);

	if(yyssh.smooth_policy == 0 && yyssh.mode == MODE_SMOOTH){
		pam_syslog(pamh, LOG_DEBUG, "smooth_policy is empty, so mode set %d", MODE_EVERY_TIME);
		yyssh.mode = MODE_EVERY_TIME;
	}

	yyssh.update_record = true;
	mode = yyssh.mode;

	if(mode == MODE_DISABLE) {
	
		r = PAM_SUCCESS;
	
	}else if(mode == MODE_EVERY_TIME) {
	
		r = send_verify(pamh);

	}else if(mode == MODE_SMOOTH) {
	
		r = check_smooth_policy(pamh);
	
	}else {
	
		pam_syslog(pamh, LOG_DEBUG, "read user:%s mode:%d not vaild.", 
				yyssh.user, mode);
		r = PAM_SYSTEM_ERR;
	}

	pam_syslog(pamh, LOG_DEBUG, "Authentication strategy %s mode:%d return:%d", 
			yyssh.user, mode, r);

	if(yyssh.update_record)
		save_user_data(pamh);

	return r;
}

PAM_EXTERN int
pam_sm_authenticate (pam_handle_t *pamh, int flags,
		     int argc, const char **argv)
{
	return sftoken_auth("auth", pamh, flags, argc, argv);
}


PAM_EXTERN int
pam_sm_setcred (pam_handle_t *pamh, int flags,
		int argc, const char **argv)
{
	return PAM_IGNORE;
}


/* password updating functions */
PAM_EXTERN int
pam_sm_chauthtok(pam_handle_t *pamh, int flags,
		 int argc, const char **argv)
{
	if (flags & PAM_PRELIM_CHECK)
		return PAM_SUCCESS;

	return sftoken_auth("password", pamh, flags, argc, argv);
}


PAM_EXTERN int
pam_sm_acct_mgmt(pam_handle_t *pamh, int flags,
		 int argc, const char **argv)
{
	return sftoken_auth("account", pamh, flags, argc, argv);

}


PAM_EXTERN int
pam_sm_open_session(pam_handle_t *pamh, int flags,
		    int argc, const char **argv)
{
	return sftoken_auth("open_session", pamh, flags, argc, argv);
}


PAM_EXTERN int
pam_sm_close_session(pam_handle_t *pamh, int flags,
		     int argc, const char **argv)
{
	return sftoken_auth("close_session", pamh, flags, argc, argv);
}

#ifdef PAM_STATI
struct pam_module _pam_exec_modstruct = {
  "pam_yyssh",
  pam_sm_authenticate,
  pam_sm_setcred,
  pam_sm_acct_mgmt,
  pam_sm_open_session,
  pam_sm_close_session,
  pam_sm_chauthtok,
};
#endif
